{ File   : NISExt.dpr
  Author : Deniz Oezmen
  Created: 2007-01-21
  Changed: 2010-08-29

  The Last Express NIS extractor/converter.
}
program NISExt;

{$APPTYPE CONSOLE}

{$R '..\res\version.res' '..\res\version.rc'}

{$DEFINE NIS_VERBOSE_PARSING_INFO}
{$DEFINE NIS_VERBOSE_PARSING_KNOWN}
{$DEFINE NIS_VERBOSE_PARSING_UNKNOWN}
{$DEFINE OVERLAY_FRAMES}

uses
  SysUtils,
  FrameDec  in 'graphics\FrameDec.pas',
  SaveTGA   in 'graphics\SaveTGA.pas',
  IMAHeader in 'sound\IMAHeader.pas',
  SoundDec  in 'sound\SoundDec.pas';

const
  ChunkType_AudioInfo         = $0003;
  ChunkType_BackgroundFrameA  = $000a;
  ChunkType_SelectBackgroundA = $000b;
  ChunkType_BackgroundFrameC  = $000c;
  ChunkType_SelectBackgroundC = $000d;
  ChunkType_OverlayFrame      = $0014;
  ChunkType_AudioData         = $0020;
  ChunkType_AudioEnd          = $0063;

type
  TNISHeader = packed record
    NumChunks: Longword;
  end;

  TNISEntry = packed record
    ChunkType: Word;
    Tag      : Word;
    ChunkSize: Longword;
  end;

  TNISEntries = packed array of TNISEntry;

function OpenInFile(var InFile: file; const InName: AnsiString): Boolean;
begin
  FileMode := fmOpenRead;
  AssignFile(InFile, InName);

  {$I-}
  Reset(InFile, 1);
  {$I+}
  Result := (IOResult = 0)
end;

function OpenOutFile(var OutFile: file; const OutName: AnsiString): Boolean;
begin
  FileMode := fmOpenWrite;
  AssignFile(OutFile, OutName);

  {$I-}
  Rewrite(OutFile, 1);
  {$I+}
  OpenOutFile := (IOResult = 0)
end;

procedure ProcessFrame(NISName: AnsiString; FrameNum: Longint;
  FrameMode: TFrameMode; const Buffer: array of Byte);
begin
  if not DumpFrame(Buffer, 0, FrameType_NIS, FrameMode,
    Format('%s_%.4d', [ExtractFileName(NISName), FrameNum])) then
    WriteLn(Format('    %s', [FrameDec.LastError]))
end;

procedure ProcessSND(NISName: AnsiString);
var
  LNKName: AnsiString;
  LNKFile: file;
  Buffer : array of Byte;
begin
  LNKName := ChangeFileExt(NISName, '.lnk');

  WriteLn(Format('  trying to link audio %s', [ExtractFileName(LNKName)]));
  if not OpenInFile(LNKFile, LNKName) then
    WriteLn('    failed: not accessible')
  else
  begin
    SetLength(Buffer, FileSize(LNKFile));
    {$I-}
    BlockRead(LNKFile, Buffer[0], Length(Buffer));
    {$I+}
    if IOResult <> 0 then
      WriteLn('    failed: read error')
    else
    begin
      LinkSNDBlocks(Buffer);

      WriteLn('    succeeded')
    end;

    CloseFile(LNKFile);
  end;

  if not DumpSND(ExtractFileName(NISName)) then
    WriteLn(Format('    %s', [SoundDec.LastError]))
end;

procedure Process(var NISFile: file; const NISName: AnsiString);
var
  i,
  FrameNum  : Longint;
  {$IFDEF OVERLAY_FRAMES}
  CurBackgr : Longint;
  {$ENDIF}
  NISHeader : TNISHeader;
  NISEntries: TNISEntries;
  Buffer    : array of Byte;
begin
  {$I-}
  BlockRead(NISFile, NISHeader, SizeOf(NISHeader));
  SetLength(NISEntries, NISHeader.NumChunks);
  BlockRead(NISFile, NISEntries[0], Length(NISEntries) * SizeOf(TNISEntry));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  not a valid NIS file');

    SetLength(NISEntries, 0);
    Exit
  end;

  FrameNum := 0;
  {$IFDEF OVERLAY_FRAMES}
  CurBackgr := 0;
  {$ENDIF}

  for i := 0 to High(NISEntries) do
    with NISEntries[i] do
    begin
      // handle informational chunks: an exact predicate for differentiation is
      // not known ...
      if ChunkType in [$0001, $0002, $0004, $0005, $000b, $000d, $0015, $0016,
        ChunkType_AudioInfo, ChunkType_AudioEnd] then
      begin
        case ChunkType of
          ChunkType_AudioInfo:
            begin
              {$IFDEF NIS_VERBOSE_PARSING_INFO}
              {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
              WriteLn(Format('  audio info: %d blocks, tag %d',
                [ChunkSize, Tag]));
              {$ENDIF}
              {$ENDIF}
              ClearSND
            end;
          ChunkType_SelectBackgroundA:
            begin
              {$IFDEF NIS_VERBOSE_PARSING_INFO}
              {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
              WriteLn('  frame info: select background A');
              {$ENDIF}
              {$ENDIF}
              {$IFDEF OVERLAY_FRAMES}
              CurBackgr := ChunkType_SelectBackgroundA
              {$ENDIF}
            end;
          ChunkType_SelectBackgroundC:
            begin
              {$IFDEF NIS_VERBOSE_PARSING_INFO}
              {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
              WriteLn('  frame info: select background C');
              {$ENDIF}
              {$ENDIF}
              {$IFDEF OVERLAY_FRAMES}
              CurBackgr := ChunkType_SelectBackgroundC
              {$ENDIF}
            end;
          ChunkType_AudioEnd:
            begin
              {$IFDEF NIS_VERBOSE_PARSING_INFO}
              {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
              WriteLn(Format('  audio end: %d blocks', [Tag]));
              {$ENDIF}
              {$ENDIF}
              TruncSND(Tag);
              ProcessSND(NISName)
            end
        {$IFDEF NIS_VERBOSE_PARSING_UNKNOWN}
        else
          WriteLn(Format('  info chunk: type 0x%.4x (size %d, tag %' +
            'd)', [ChunkType, ChunkSize, Tag]))
        {$ENDIF}
        end;

        // despite their size field, info chunks don't have a payload
        Continue
      end;

      // read chunk payload
      SetLength(Buffer, ChunkSize);
      {$I-}
      BlockRead(NISFile, Buffer[0], ChunkSize);
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    error reading chunk data');
        Break
      end;

      case ChunkType of
        ChunkType_AudioData:
          begin
            {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
            WriteLn(Format('  audio (%d blocks, %d bytes, tag %d)',
              [ChunkSize div SoundDec.SNDBlockSize, ChunkSize, Tag]));
            {$ENDIF}
            AddSNDBlocks(Buffer)
          end;
        ChunkType_BackgroundFrameA:
          begin
            {$IFDEF OVERLAY_FRAMES}
            {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
            WriteLn(Format('  frame (background type 0x%.4x, %d bytes, tag %d)',
              [ChunkType, ChunkSize, Tag]));
            {$ENDIF}
            ProcessFrame(NISName, FrameNum, FrameMode_BackgroundA, Buffer);
            {$ELSE}
            {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
            WriteLn(Format('  frame #%.4d (background type 0x%.4x, %d bytes, ' +
              'tag %d)', [FrameNum, ChunkType, ChunkSize, Tag]));
            {$ENDIF}
            ProcessFrame(NISName, FrameNum, FrameMode_Single, Buffer);
            Inc(FrameNum)
            {$ENDIF}
          end;
        ChunkType_BackgroundFrameC:
          begin
            {$IFDEF OVERLAY_FRAMES}
            {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
            WriteLn(Format('  frame (background type 0x%.4x, %d bytes, tag %d)',
              [ChunkType, ChunkSize, Tag]));
            {$ENDIF}
            ProcessFrame(NISName, FrameNum, FrameMode_BackgroundC, Buffer);
            {$ELSE}
            {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
            WriteLn(Format('  frame #%.4d (background type 0x%.4x, %d bytes, ' +
              'tag %d)', [FrameNum, ChunkType, ChunkSize, Tag]));
            {$ENDIF}
            ProcessFrame(NISName, FrameNum, FrameMode_Single, Buffer);
            Inc(FrameNum)
            {$ENDIF}
          end;
        ChunkType_OverlayFrame:
          begin
            {$IFDEF NIS_VERBOSE_PARSING_KNOWN}
            WriteLn(Format('  frame #%.4d (overlay, %d bytes, tag %d)',
              [FrameNum, ChunkSize, Tag]));
            {$ENDIF}
            {$IFDEF OVERLAY_FRAMES}
            if CurBackgr = ChunkType_SelectBackgroundA then
              ProcessFrame(NISName, FrameNum, FrameMode_OverlayA, Buffer)
            else if CurBackgr = ChunkType_SelectBackgroundC then
              ProcessFrame(NISName, FrameNum, FrameMode_OverlayC, Buffer)
            else
              ProcessFrame(NISName, FrameNum, FrameMode_Single, Buffer);
            {$ELSE}
            ProcessFrame(NISName, FrameNum, FrameMode_Single, Buffer);
            {$ENDIF}
            Inc(FrameNum)
          end
      {$IFDEF NIS_VERBOSE_PARSING_UNKNOWN}
      else
        WriteLn(Format('  unsupported chunk: type 0x%.4x (%d bytes, tag %d, o' +
          'ffset %d)',
          [ChunkType, ChunkSize, Tag, FilePos(NISFile) - Longint(ChunkSize)]))
      {$ENDIF}
      end
    end;

  SetLength(Buffer, 0)
end;

procedure ScanFiles(FileParam: AnsiString);
var
  Dir,
  InName   : AnsiString;
  InFile   : file;
  SearchRec: TSearchRec;
begin
  // get the directory information from FileParam
  Dir := ExtractFilePath(FileParam);

  // search for all files matching FileParam
  if FindFirst(FileParam, faAnyFile - faDirectory, SearchRec) = 0 then
  begin
    repeat
      InName := Dir + SearchRec.Name;

      WriteLn(InName);

      // if this file is accessible, analyze it
      if OpenInFile(InFile, InName) then
      begin
        Process(InFile, InName);
        Close(InFile)
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [FileParam]));

  FindClose(SearchRec)
end;

begin
  if not SoundDec.Initialized then
  begin
    WriteLn(Format('Error: Could not initialize library %s.',
      [IMAHeader.DLLName]));
    Exit
  end;

  if ParamCount = 1 then
    ScanFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.


