{ File   : SoundDec.dpr
  Author : Deniz Oezmen
  Created: 2009-03-11, based on NISExt.dpr created 2007-01-21
  Changed: 2010-08-29

  The Last Express audio decompression helper module.
}
unit SoundDec;

interface

const
  SNDBlockSize = 739;

var
  LastError  : AnsiString;
  Initialized: Boolean;

procedure ClearSND;
procedure AddSNDBlocks(const Buffer: array of Byte; BaseIdx: Longint = 0);
procedure LinkSNDBlocks(const Buffer: array of Byte);
procedure TruncSND(NumBlocks: Longint);
function DumpSND(SNDName: AnsiString): Boolean;

implementation

uses
  SysUtils,
  IMAHeader;

type
  TSNDHeader = packed record
    Size     : Longint;  // NIS: size of all blocks, including those located in
                         //      the matching LNK file
                         // LNK: size of the LNK file itself, including the
                         //      header
                         // SND: size of all blocks
    NumBlocks: Word;     // NIS/LNK/SND: # of blocks available in current file
  end;

  TWAVHeader = packed record
     RIFFHeader: packed record
       GroupID : array[0..3] of Char;
       Size    : Longint;
       RiffType: array[0..3] of Char;
     end;
     FormatChunk     : packed record
       ChunkID       : array[0..3] of Char;
       ChunkSize     : Longint;
       FormatTag     : Word;
       Channels      : Word;
       SamplesPerSec : Longint;
       AvgBytesPerSec: Longint;
       BlockAlign    : Word;
       BitsPerSample : Word;
     end;
     DataChunk: packed record
       ChunkID  : array[0..3] of Char;
       ChunkSize: Longint;
     end;
  end;

var
  SNDBuf   : array of Byte;
  FreshSND,
  LinkedSND: Boolean;
  SNDHeader: TSNDHeader;
  WAVHeader: TWAVHeader = (RIFFHeader:
                            (GroupID: 'RIFF';
                             RiffType: 'WAVE');
                           FormatChunk:
                             (ChunkID: 'fmt ';
                              ChunkSize: 16;
                              FormatTag: 1;
                              Channels: 1;
                              SamplesPerSec: 44100;
                              AvgBytesPerSec: 88200;
                              BlockAlign: 2;
                              BitsPerSample: 16);
                           DataChunk:
                             (ChunkID: 'data'));

procedure TruncSND(NumBlocks: Longint);
var
  Size: Longint;
begin
  Size := NumBlocks * SNDBlockSize;

  if Length(SNDBuf) > Size then
    SetLength(SNDBuf, Size)
end;

procedure ClearSND;
begin
  TruncSND(0);
  FreshSND := True;
  LinkedSND := False
end;

procedure AddSNDBlocks(const Buffer: array of Byte; BaseIdx: Longint = 0);
var
  Idx,
  Len: Longint;
begin
  // if this is a "fresh" SND, we expect to see a header first
  if FreshSND then
  begin
    Move(Buffer[BaseIdx], SNDHeader, SizeOf(SNDHeader));
    Idx := BaseIdx + SizeOf(SNDHeader)
  end
  else
    Idx := BaseIdx;

  Len := Length(Buffer) - Idx;

  SetLength(SNDBuf, Length(SNDBuf) + Len);
  Move(Buffer[Idx], SNDBuf[Length(SNDBuf) - Len], Len);

  // truncate if necessary
  TruncSND(SNDHeader.NumBlocks);

  FreshSND := False
end;

procedure LinkSNDBlocks(const Buffer: array of Byte);
var
  LNKHeader: TSNDHeader;
begin
  Move(Buffer[0], LNKHeader, SizeOf(LNKHeader));
  Inc(SNDHeader.NumBlocks, LNKHeader.NumBlocks);

  AddSNDBlocks(Buffer, SizeOf(LNKHeader));

  LinkedSND := True
end;

function DumpSND(SNDName: AnsiString): Boolean;
var
  Handle,
  UncompSize: Longint;
  SampleBuf : array of Smallint;
  WAVFile   : file;
begin
  Result := False;

  // open output wave file
  AssignFile(WAVFile, Format('%s.wav', [SNDName]));
  {$I-}
  Rewrite(WAVFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    LastError := 'error opening output';
    Exit
  end;

  Handle := GetIMAManager(SNDBlockSize, 0, 1, IMA_ADPCM_Format_Express);
  if Handle < Handler_Success then
  begin
    LastError := Format('audio manager error: %s' , [GetErrorName(Handle)]);

    Close(WAVFile);
    Exit
  end;

  IMALoadCodes(Handle, SNDBuf[0], Length(SNDBuf));

  SetLength(SampleBuf, IMAGetDecompressedSize(Handle));
  IMADecompress(Handle, SampleBuf[0]);

  IMAUnloadCodes(Handle);
  ReleaseIMAManager(Handle);

  // prepare wave header
  UncompSize := Length(SampleBuf) * 2;

  WAVHeader.RIFFHeader.Size := UncompSize + 32;
  WAVHeader.DataChunk.ChunkSize := UncompSize;

  // output wave data
  {$I-}
  BlockWrite(WAVFile, WAVHeader, SizeOf(WAVHeader));
  BlockWrite(WAVFile, SampleBuf[0], Length(SampleBuf) * 2);
  {$I+}
  Result := (IOResult = 0);

  CloseFile(WAVFile);

  if not Result then
    LastError := 'write error'
  else
    LastError := '';

  if Result and LinkedSND then
  begin
     Result := SNDHeader.NumBlocks * SNDBlockSize = Length(SNDBuf);
     if not Result then
       LastError := Format('size mismatch: expected %d bytes, got %d bytes',
         [SNDHeader.NumBlocks * SNDBlockSize, Length(SNDBuf)])
  end;

  ClearSND;
  SetLength(SampleBuf, 0)
end;

initialization
  Initialized := LoadDLL(ExtractFilePath(ParamStr(0)));

  if Initialized then
    Initialized := DLLCanUse;

finalization
  UnloadDLL;

end.
