{ File   : FrameDec.dpr
  Author : Deniz Oezmen
  Created: 2009-03-05, based on NISExt.dpr created 2007-01-21
  Changed: 2009-10-24

  The Last Express frame decompression helper module.
}
unit FrameDec;

{$DEFINE HIDELINEPARITIES}
{$DEFINE FILLCHECKERBOARD}

interface

type
  TFrameType = (FrameType_SEQ, FrameType_NIS);
  TFrameMode = (FrameMode_BackgroundA, FrameMode_BackgroundC,
                FrameMode_OverlayA, FrameMode_OverlayC,
                FrameMode_Single);

var
  LastError: AnsiString;

function DumpFrame(const Buffer: array of Byte; BaseIdx: Longint;
  FrameType: TFrameType; FrameMode: TFrameMode;
  const FrameName: AnsiString): Boolean;

implementation

uses
  SysUtils,
  SaveTGA;

const
  ScreenWidth    = 640;
  ScreenHeight   = 480;
  MaxPaletteSize = 256;
  TransCol       = 0;
  LineParBaseCol = MaxPaletteSize - 2;

  CompPos_SEQ    = $002c;
  CompPos_NIS    = $0124;

type
  TTLEFrame = packed record
    DataOfs    : Longword;
    Unknown1   : Longword;
    PalOfs     : Longword;
    XPos1      : Longword;
    Unknown2   : Longword;
    XPos2      : Longword;
    Unknown3   : Longword;
    InitialSkip: Longword; // doubled, since each pixel occupies one color word
    DecompSize : Longword; // as above
    // NIS frame headers end here. SEQ frame headers have additional 32 bytes of
    // data, notably the compression type at the position outlined above in
    // CompPos_SEQ
  end;
  PTLEFrame = ^TTLEFrame;

  TTLEPal = array[0..MaxPaletteSize - 1] of Word;
  PTLEPal = ^TTLEPal;

  TScreen = array[0..ScreenWidth * ScreenHeight - 1] of Byte;
  PScreen = ^TScreen;

  TLargeScreen = array[0..ScreenWidth * ScreenHeight - 1] of Word;

var
  BackgroundA,
  BackgroundC: TScreen;
  PalA       : TPalette4;
  PalC       : TPalette4;

procedure ConvertPal(Pal: PTLEPal; PalSize: Longint; out Palette: TPalette4);
var
  i: Longint;
begin
  // initialize whole palette using pure (transparent) green to make detection
  // of errors easier
  SetLength(Palette, MaxPaletteSize);
  for i := 0 to High(Palette) do
    with Palette[i] do
    begin
      B := 0;
      G := 255;
      R := 0;
      A := 0
    end;

  {$IFNDEF HIDELINEPARITIES}
  // if line parities are to be visualized, use the upper two palette entries
  // for the special colors -- these are unlikely (impossible?) to be occupied
  // otherwise
  with Palette[LineParBaseCol] do
  begin
    B := 255;
    G := 0;
    R := 255;
    A := 255
  end;

  with Palette[LineParBaseCol + 1] do
  begin
    B := 0;
    G := 0;
    R := 255;
    A := 255
  end;
  {$ENDIF}

  // convert actual colors
  for i := 0 to PalSize - 1 do
    with Palette[i] do
    begin
      B :=  (Pal^[i]          and $1f) shl 3;
      G := ((Pal^[i] shr   5) and $1f) shl 3;
      R := ((Pal^[i] shr  10) and $1f) shl 3;
      
      if i = TransCol then
        A := 0
      else
        A := 255
    end
end;

function Decompress_Empty(TLEFrame: PTLEFrame; const TLEBuffer: array of Byte;
  Screen: PScreen; out PalSize: Longint): Boolean;
begin
  Result := False;
  PalSize := 0;

  if TLEFrame = nil then
    Exit;

  // clear output buffer
  FillChar(Screen^[0], Length(Screen^), TransCol);
  PalSize := TransCol + 1;

  Result := True
end;

// expressw.exe .text:00425C72
function Decompress_03(TLEFrame: PTLEFrame; const TLEBuffer: array of Byte;
  Screen: PScreen; out PalSize: Longint): Boolean;
var
  i,
  NumBlanks,
  InIdx,
  OutIdx,
  OpCode    : Longint;
  Value,
  LineParity: Byte;
begin
  Result := False;
  PalSize := 0;

  if TLEFrame = nil then
    Exit;

  // clear output buffer
  FillChar(Screen^[0], Length(Screen^), TransCol);
  PalSize := TransCol + 1;

  // default number of columns to skip
  NumBlanks := (ScreenWidth - 1) - (TLEFrame^.XPos2 - TLEFrame^.XPos1);

  InIdx := TLEFrame^.DataOfs;
  OutIdx := TLEFrame^.InitialSkip shr 1;

  // initialize parity
  LineParity := (OutIdx div ScreenWidth) and $01;

  while OutIdx < Longint(TLEFrame^.DecompSize shr 1) do
  begin
    if InIdx > High(TLEBuffer) then
      Exit;

    OpCode := TLEBuffer[InIdx];
    Inc(InIdx);

    if OpCode and $80 <> 0 then
    begin
      if OpCode and $40 <> 0 then
      begin
        OpCode := OpCode and $3f;

        {$IFNDEF HIDELINEPARITIES}
        if OutIdx and $01 <> LineParity then
        begin
          if OutIdx > High(Screen^) then
            Exit;

          if PalSize < 2 then
            PalSize := 2;

          Screen^[OutIdx] := 1
        end;
        {$ENDIF}

        Inc(OutIdx, NumBlanks + OpCode + 1);

        LineParity := (OutIdx div ScreenWidth) and $01;

        {$IFNDEF HIDELINEPARITIES}
        if OutIdx and $01 <> LineParity then
        begin
          if OutIdx > High(Screen^) then
            Exit;

          Screen^[OutIdx] := LineParBaseCol + LineParity
        end;
        {$ENDIF}

        Inc(OutIdx)
      end
      else
      begin
        OpCode := OpCode and $3f;

        if OpCode and $20 <> 0 then
        begin
          if InIdx > High(TLEBuffer) then
            Exit;

          OpCode := (OpCode and $1f) shl 8 + TLEBuffer[InIdx];
          Inc(InIdx);

          if OpCode and $1000 <> 0 then
          begin
            Inc(OutIdx, OpCode and $0fff);
            LineParity := (OutIdx div ScreenWidth) and $01;

            Continue
          end
        end;

        begin
          {$IFNDEF HIDELINEPARITIES}
          if OutIdx and $01 <> LineParity then
          begin
            if OutIdx > High(Screen^) then
              Exit;

            if PalSize < 2 then
              PalSize := 2;

            Screen^[OutIdx] := 1
          end;
          {$ENDIF}

          Inc(OutIdx, OpCode + 1);

          LineParity := (OutIdx div ScreenWidth) and $01;

          {$IFNDEF HIDELINEPARITIES}
          if OutIdx and $01 <> LineParity then
          begin
            if OutIdx > High(Screen^) then
              Exit;

            Screen^[OutIdx] := LineParBaseCol + LineParity
          end;
          {$ENDIF}

          Inc(OutIdx)
        end
      end
    end
    else
    begin
      Value := OpCode and $07;
      OpCode := OpCode shr 3;

      if PalSize < Value + 1 then
        PalSize := Value + 1;

      if OpCode = 0 then
      begin
        if InIdx > High(TLEBuffer) then
          Exit;

        OpCode := TLEBuffer[InIdx];
        Inc(InIdx)
      end;

      if OutIdx and $01 = LineParity then
      begin
        {$IFDEF FILLCHECKERBOARD}
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;
        {$ENDIF}

        Inc(OutIdx);
        Dec(OpCode);

        if OpCode = 0 then
          Continue
      end;

      for i := 1 to OpCode shr 1 do
      begin
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;

        {$IFDEF FILLCHECKERBOARD}
        if OutIdx + 1 > High(Screen^) then
          Exit;

        Screen^[OutIdx + 1] := Value;
        {$ENDIF}

        Inc(OutIdx, 2)
      end;

      if OpCode and $01 <> 0 then
      begin
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;
        Inc(OutIdx)
      end
    end
  end;

  Result := True
end;

// expressw.exe .text:004257CC
function Decompress_04(TLEFrame: PTLEFrame; const TLEBuffer: array of Byte;
  Screen: PScreen; out PalSize: Longint): Boolean;
var
  i,
  NumBlanks,
  InIdx,
  OutIdx,
  OpCode    : Longint;
  Value,
  LineParity: Byte;
begin
  Result := False;
  PalSize := 0;

  if TLEFrame = nil then
    Exit;

  // clear output buffer
  FillChar(Screen^[0], Length(Screen^), TransCol);
  PalSize := TransCol + 1;

  // default number of columns to skip
  NumBlanks := (ScreenWidth - 1) - (TLEFrame^.XPos2 - TLEFrame^.XPos1);

  InIdx := TLEFrame^.DataOfs;
  OutIdx := TLEFrame^.InitialSkip shr 1;

  // initialize parity
  LineParity := (OutIdx div ScreenWidth) and $01;

  while OutIdx < Longint(TLEFrame^.DecompSize shr 1) do
  begin
    if InIdx > High(TLEBuffer) then
      Exit;

    OpCode := TLEBuffer[InIdx];
    Inc(InIdx);

    if OpCode and $80 <> 0 then
    begin
      if OpCode and $40 <> 0 then
      begin
        OpCode := OpCode and $3f;

        {$IFNDEF HIDELINEPARITIES}
        if OutIdx and $01 <> LineParity then
        begin
          if OutIdx > High(Screen^) then
            Exit;

          if PalSize < 2 then
            PalSize := 2;

          Screen^[OutIdx] := 1
        end;
        {$ENDIF}

        Inc(OutIdx, NumBlanks + OpCode + 1);

        LineParity := (OutIdx div ScreenWidth) and $01;

        {$IFNDEF HIDELINEPARITIES}
        if OutIdx and $01 <> LineParity then
        begin
          if OutIdx > High(Screen^) then
            Exit;

          Screen^[OutIdx] := LineParBaseCol + LineParity
        end;
        {$ENDIF}

        Inc(OutIdx)
      end
      else
      begin
        OpCode := OpCode and $3f;

        if OpCode and $20 <> 0 then
        begin
          if InIdx > High(TLEBuffer) then
            Exit;

          OpCode := (OpCode and $1f) shl 8 + TLEBuffer[InIdx];
          Inc(InIdx);

          if OpCode and $1000 <> 0 then
          begin
            Inc(OutIdx, OpCode and $0fff);
            LineParity := (OutIdx div ScreenWidth) and $01;

            Continue
          end
        end;

        begin
          {$IFNDEF HIDELINEPARITIES}
          if OutIdx and $01 <> LineParity then
          begin
            if OutIdx > High(Screen^) then
              Exit;

            if PalSize < Value + 1 then
              PalSize := Value + 1;

            Screen^[OutIdx] := 1
          end;
          {$ENDIF}

          Inc(OutIdx, OpCode + 1);

          LineParity := (OutIdx div ScreenWidth) and $01;

          {$IFNDEF HIDELINEPARITIES}
          if OutIdx and $01 <> LineParity then
          begin
            if OutIdx > High(Screen^) then
              Exit;

            Screen^[OutIdx] := LineParBaseCol + LineParity
          end;
          {$ENDIF}

          Inc(OutIdx)
        end
      end
    end
    else
    begin
      Value := OpCode and $0f;
      OpCode := OpCode shr 4;

      if PalSize < Value + 1 then
        PalSize := Value + 1;

      if OpCode = 0 then
      begin
        if InIdx > High(TLEBuffer) then
          Exit;

        OpCode := TLEBuffer[InIdx];
        Inc(InIdx)
      end;

      if OutIdx and $01 = LineParity then
      begin
        {$IFDEF FILLCHECKERBOARD}
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;
        {$ENDIF}

        Inc(OutIdx);
        Dec(OpCode);

        if OpCode = 0 then
          Continue
      end;

      for i := 1 to OpCode shr 1 do
      begin
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;

        {$IFDEF FILLCHECKERBOARD}
        if OutIdx + 1 > High(Screen^) then
          Exit;

        Screen^[OutIdx + 1] := Value;
        {$ENDIF}

        Inc(OutIdx, 2)
      end;

      if OpCode and $01 <> 0 then
      begin
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;
        Inc(OutIdx)
      end
    end
  end;

  Result := True
end;

// expressw.exe .text:004253E5
function Decompress_05(TLEFrame: PTLEFrame; const TLEBuffer: array of Byte;
  Screen: PScreen; out PalSize: Longint): Boolean;
var
  i,
  InIdx,
  OutIdx,
  OpCode    : Longint;
  Value,
  LineParity: Byte;
begin
  Result := False;
  PalSize := 0;

  if TLEFrame = nil then
    Exit;

  // clear output buffer
  FillChar(Screen^[0], Length(Screen^), TransCol);
  PalSize := TransCol + 1;

  InIdx := TLEFrame^.DataOfs;
  OutIdx := TLEFrame^.InitialSkip shr 1;

  // initialize parity
  LineParity := (OutIdx div ScreenWidth) and $01;

  while OutIdx < Longint(TLEFrame^.DecompSize shr 1) do
  begin
    if InIdx > High(TLEBuffer) then
      Exit;

    OpCode := TLEBuffer[InIdx];
    Inc(InIdx);

    if OpCode and $1f = 0 then
    begin
      if InIdx > High(TLEBuffer) then
        Exit;

      OpCode := OpCode shl 3 + TLEBuffer[InIdx];
      Inc(InIdx);

      if OpCode and $0400 <> 0 then
      begin
        Inc(OutIdx, OpCode and $03ff);

        LineParity := (OutIdx div ScreenWidth) and $01
      end
      else
      begin
        {$IFNDEF HIDELINEPARITIES}
        if OutIdx and $01 <> LineParity then
        begin
          if OutIdx > High(Screen^) then
            Exit;

          if PalSize < 2 then
            PalSize := 2;

          Screen^[OutIdx] := 1
        end;
        {$ENDIF}

        Inc(OutIdx, OpCode + 1);

        LineParity := (OutIdx div ScreenWidth) and $01;

        {$IFNDEF HIDELINEPARITIES}
        if OutIdx and $01 <> LineParity then
        begin
          if OutIdx > High(Screen^) then
            Exit;

          Screen^[OutIdx] := LineParBaseCol + LineParity
        end;
        {$ENDIF}

        Inc(OutIdx)
      end
    end
    else
    begin
      Value := OpCode and $1f;
      OpCode := OpCode shr 5;

      if PalSize < Value + 1 then
        PalSize := Value + 1;

      if OpCode = 0 then
      begin
        if InIdx > High(TLEBuffer) then
          Exit;

        OpCode := TLEBuffer[InIdx];
        Inc(InIdx)
      end;

      if OutIdx and $01 = LineParity then
      begin
        {$IFDEF FILLCHECKERBOARD}
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;
        {$ENDIF}

        Inc(OutIdx);
        Dec(OpCode);

        if OpCode = 0 then
          Continue
      end;

      for i := 1 to OpCode shr 1 do
      begin
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;

        {$IFDEF FILLCHECKERBOARD}
        if OutIdx + 1 > High(Screen^) then
          Exit;

        Screen^[OutIdx + 1] := Value;
        {$ENDIF}

        Inc(OutIdx, 2)
      end;

      if OpCode and $01 <> 0 then
      begin
        if OutIdx > High(Screen^) then
          Exit;

        Screen^[OutIdx] := Value;
        Inc(OutIdx)
      end
    end
  end;

  Result := True
end;

// expressw.exe .text:0042614B
function Decompress_07(TLEFrame: PTLEFrame; const TLEBuffer: array of Byte;
  Screen: PScreen; out PalSize: Longint): Boolean;
var
  i,
  NumBlanks,
  InIdx,
  OutIdx,
  OpCode    : Longint;
  Value,
  LineParity: Byte;
begin
  Result := False;
  PalSize := 0;

  if TLEFrame = nil then
    Exit;

  // clear output buffer
  FillChar(Screen^[0], Length(Screen^), TransCol);
  PalSize := TransCol + 1;

  // default number of columns to skip
  NumBlanks := (ScreenWidth - 1) - (TLEFrame^.XPos2 - TLEFrame^.XPos1);

  InIdx := TLEFrame^.DataOfs;
  OutIdx := TLEFrame^.InitialSkip shr 1;

  // initialize parity
  LineParity := (OutIdx div ScreenWidth) and $01;

  while OutIdx < Longint(TLEFrame^.DecompSize shr 1) do
  begin
    if InIdx > High(TLEBuffer) then
      Exit;

    OpCode := TLEBuffer[InIdx];
    Inc(InIdx);

    if OpCode and $80 <> 0 then
    begin
      if OpCode and $40 <> 0 then
      begin
        if OpCode and $20 <> 0 then
        begin
          OpCode := OpCode and $1f;

          {$IFNDEF HIDELINEPARITIES}
          if OutIdx and $01 <> LineParity then
          begin
            if OutIdx > High(Screen^) then
              Exit;

            if PalSize < 2 then
              PalSize := 2;

            Screen^[OutIdx] := 1
          end;
          {$ENDIF}

          Inc(OutIdx, NumBlanks + OpCode + 1);

          LineParity := (OutIdx div ScreenWidth) and $01;

          {$IFNDEF HIDELINEPARITIES}
          if OutIdx and $01 <> LineParity then
          begin
            if OutIdx > High(Screen^) then
              Exit;

            Screen^[OutIdx] := LineParBaseCol + LineParity
          end;
          {$ENDIF}

          Inc(OutIdx)
        end
        else
        begin
          OpCode := OpCode and $1f;

          if OpCode and $10 <> 0 then
          begin
            if InIdx > High(TLEBuffer) then
              Exit;

            OpCode := (OpCode and $0f) shl 8 + TLEBuffer[InIdx];
            Inc(InIdx);

            if OpCode and $0800 <> 0 then
            begin
              Inc(OutIdx, OpCode and $07ff);
              LineParity := (OutIdx div ScreenWidth) and $01;

              Continue
            end
          end;

          begin
            {$IFNDEF HIDELINEPARITIES}
            if OutIdx and $01 <> LineParity then
            begin
              if OutIdx > High(Screen^) then
                Exit;

              if PalSize < 2 then
                PalSize := 2;

              Screen^[OutIdx] := 1
            end;
            {$ENDIF}

            Inc(OutIdx, OpCode + 1);

            LineParity := (OutIdx div ScreenWidth) and $01;

            {$IFNDEF HIDELINEPARITIES}
            if OutIdx and $01 <> LineParity then
            begin
              if OutIdx > High(Screen^) then
                Exit;

              Screen^[OutIdx] := LineParBaseCol + LineParity
            end;
            {$ENDIF}

            Inc(OutIdx)
          end
        end
      end
      else
      begin
        OpCode := OpCode and $3f;

        if InIdx > High(TLEBuffer) then
          Exit;

        Value := TLEBuffer[InIdx];
        Inc(InIdx);

        if PalSize < Value + 1 then
          PalSize := Value + 1;

        if OutIdx and $01 = LineParity then
        begin
          {$IFDEF FILLCHECKERBOARD}
          if OutIdx > High(Screen^) then
            Exit;

          Screen^[OutIdx] := Value;
          {$ENDIF}

          Inc(OutIdx);
          Dec(OpCode);

          if OpCode = 0 then
            Continue
        end;

        for i := 1 to OpCode shr 1 do
        begin
          if OutIdx > High(Screen^) then
            Exit;
            
          Screen^[OutIdx] := Value;

          {$IFDEF FILLCHECKERBOARD}
          if OutIdx + 1 > High(Screen^) then
            Exit;

          Screen^[OutIdx + 1] := Value;
          {$ENDIF}

          Inc(OutIdx, 2)
        end;

        if OpCode and $01 <> 0 then
        begin
          if OutIdx > High(Screen^) then
            Exit;

          Screen^[OutIdx] := Value;
          Inc(OutIdx)
        end
      end
    end
    else
    begin
      {$IFNDEF FILLCHECKERBOARD}
      if OutIdx and $01 <> LineParity then
      {$ENDIF}
      begin
        if OutIdx > High(Screen^) then
          Exit;

        if PalSize < OpCode + 1 then
          PalSize := OpCode + 1;

        Screen^[OutIdx] := OpCode
      end;

      Inc(OutIdx)
    end
  end;

  Result := True
end;

// expressw.exe .text:00424F10
function Decompress_ff(TLEFrame: PTLEFrame; const TLEBuffer: array of Byte;
  Screen: PScreen; out PalSize: Longint): Boolean;
var
  i,
  InIdx,
  OutIdx,
  Pos,
  Len   : Longint;
  OpCode: Word;
  Value : Byte;
begin
  Result := False;
  PalSize := 0;

  if TLEFrame = nil then
    Exit;

  // clear output buffer
  FillChar(Screen^[0], Length(Screen^), TransCol);
  PalSize := TransCol + 1;

  InIdx := TLEFrame^.DataOfs;
  OutIdx := 0;

  while OutIdx < Length(Screen^) do
  begin
    if InIdx + 1 > High(TLEBuffer) then
      Exit;

    Move(TLEBuffer[InIdx], OpCode, 2);

    if OpCode and $00ff < $80 then
    begin
      Inc(InIdx);

      Value := OpCode and $00ff;

      if PalSize < Value + 1 then
        PalSize := Value + 1;

      if OutIdx > High(Screen^) then
        Exit;

      Screen^[OutIdx] := Value;
      Inc(OutIdx);
    end
    else
    begin
      Inc(InIdx, 2);

      if OpCode and $00ff < $f0 then
      begin
        if OpCode and $00ff < $e0 then
        begin
          Pos := OutIdx + (((OpCode and $0007) shl 8) or (OpCode shr 8)) - 2048;
          Len := (OpCode and $0078) shr 3 + 3;

          if ((Pos + Len - 1) > High(Screen^)) or
            ((OutIdx + Len - 1) > High(Screen^)) then
            Exit;

          for i := 1 to Len do
          begin
            Screen^[OutIdx] := Screen^[Pos];
            Inc(Pos);
            Inc(OutIdx)
          end
        end
        else
        begin
          Value := OpCode shr 8;

          if PalSize < Value + 1 then
            PalSize := Value + 1;

          Len := OpCode and $0f + 1;

          if OutIdx + Len - 1 > High(Screen^) then
            Exit;

          for i := 1 to Len do
          begin
            Screen^[OutIdx] := Value;
            Inc(OutIdx)
          end
        end
      end
      else
        Inc(OutIdx, (((OpCode and $00ff) shl 8) or (OpCode shr 8)) and $0fff)
    end
  end;

  Result := True
end;

function OverlayAndDumpFrame(Overlay, Background: PScreen; OverlayPal,
  BackgroundPal: TPalette4; const FrameName: AnsiString): Boolean;
var
  i          : Longint;
  LargeScreen: TLargeScreen;
  Palette    : TPalette4;
begin
  // overlay frames
  for i := 0 to High(LargeScreen) do
    if Overlay^[i] = TransCol then
      LargeScreen[i] := Background^[i]
    else
      LargeScreen[i] := Overlay^[i] + MaxPaletteSize;

  // mix palettes
  SetLength(Palette, 2 * MaxPaletteSize);
  Move(BackgroundPal[0], Palette[0], MaxPaletteSize * SizeOf(Palette[0]));
  Move(OverlayPal[0], Palette[MaxPaletteSize],
    MaxPaletteSize * SizeOf(Palette[0]));

  // dump
  Result := DumpIndexedTGA(Format('%s.tga', [FrameName]), ScreenWidth,
    ScreenHeight, @LargeScreen[0], @Palette[0], Length(Palette),
    TGAPixelDepth_32Bit, TGACompression_RLE)
end;

function DumpFrame(const Buffer: array of Byte; BaseIdx: Longint;
  FrameType: TFrameType; FrameMode: TFrameMode;
  const FrameName: AnsiString): Boolean;
var
  Palette : TPalette4;
  BufIdx,
  PalSize : Longint;
  TLEFrame: PTLEFrame;
  TLEPal  : PTLEPal;
  Screen  : TScreen;
  CompType: Byte;
  CompRes : Boolean;
begin
  Result := False;

  if Length(Buffer) < SizeOf(TTLEFrame) then
  begin
    LastError := 'invalid frame';
    Exit
  end;

  // get frame start
  TLEFrame := @Buffer[BaseIdx];

  // get compression type
  case FrameType of
    FrameType_SEQ:
      BufIdx := BaseIdx + CompPos_SEQ;
    FrameType_NIS:
      BufIdx := BaseIdx + CompPos_NIS
  else
    LastError := Format('invalid frame type specifier %d',
      [Longint(FrameType)]);
    Exit
  end;

  if BufIdx > High(Buffer) then
  begin
    LastError := Format('cannot determine compression type at offset 0x%.8x',
      [BufIdx]);
    Exit
  end;

  CompType := Buffer[BufIdx];

  // does the image fit into the screen?
  if Longint(TLEFrame^.DecompSize shr 1) > Length(Screen) then
  begin
    LastError := 'image too large';
    Exit
  end;

  // could this frame be empty?
  BufIdx := TLEFrame^.DataOfs;
  if (BufIdx = 0) or (BufIdx > High(Buffer)) then
    CompRes := Decompress_Empty(TLEFrame, Buffer, @Screen[0], PalSize)
  else
  begin
    // decompress image
    case CompType of
      $03: CompRes := Decompress_03(TLEFrame, Buffer, @Screen[0], PalSize);
      $04: CompRes := Decompress_04(TLEFrame, Buffer, @Screen[0], PalSize);
      $05: CompRes := Decompress_05(TLEFrame, Buffer, @Screen[0], PalSize);
      $07: CompRes := Decompress_07(TLEFrame, Buffer, @Screen[0], PalSize);
      $ff: CompRes := Decompress_ff(TLEFrame, Buffer, @Screen[0], PalSize)
    else
      LastError := Format('unsupported compression type 0x%.2x', [CompType]);
      Exit
    end
  end;

  if not CompRes then
  begin
    LastError := Format('decompression error (type %.2x)', [CompType]);
    Exit
  end;

  // are we prepared to address this palette?
  if PalSize > MaxPaletteSize then
  begin
    LastError := Format('palette size %d too large', [PalSize]);
    Exit
  end;

  // is the buffer able to store the indicated palette?
  BufIdx := TLEFrame^.PalOfs;
  if BufIdx > Length(Buffer) - 2 * PalSize then
  begin
    LastError := Format('invalid palette reference to offset 0x%.8x, %d entri' +
      'es', [BufIdx, PalSize]);
    Exit
  end;

  // convert palette
  TLEPal := @Buffer[BufIdx];
  ConvertPal(TLEPal, PalSize, Palette);

  // output final image: handle frame mode
  case FrameMode of
    FrameMode_BackgroundA:
      begin
        Move(Screen[0], BackgroundA[0], Length(Screen));
        Move(Palette[0], PalA[0], MaxPaletteSize * SizeOf(Palette[0]));
        Result := True
      end;
    FrameMode_BackgroundC:
      begin
        Move(Screen[0], BackgroundC[0], Length(Screen));
        Move(Palette[0], PalC[0], MaxPaletteSize * SizeOf(Palette[0]));
        Result := True
      end;
    FrameMode_OverlayA:
      Result := OverlayAndDumpFrame(@Screen[0], @BackgroundA[0], Palette, PalA,
        FrameName);
    FrameMode_OverlayC:
      Result := OverlayAndDumpFrame(@Screen[0], @BackgroundC[0], Palette, PalC,
        FrameName);
    FrameMode_Single:
      Result := DumpIndexedTGA(Format('%s.tga', [FrameName]), ScreenWidth,
        ScreenHeight, @Screen[0], @Palette[0], Length(Palette),
        TGAPixelDepth_32Bit, TGACompression_RLE)
  end;

  if not Result then
    LastError := 'dump failed'
  else
    LastError := '';

  SetLength(Palette, 0)
end;

initialization
  SetLength(PalA, MaxPaletteSize * 2);
  SetLength(PalC, MaxPaletteSize * 2);

finalization
  SetLength(PalA, 0);
  SetLength(PalC, 0);

end.
